//=============================================================================
// VotingHandler. Coded by .:..: (Ideas from UT2004)
//=============================================================================
class VotingHandler expands Mutator
	Config(VotingManager);

// Assume there are never more than 32 players in 1 server
Const MaxUnrealPlayers=32;

var VotingReplicationInfo ActiveVReplications[32];
var PlayerPawn UpdatingPlayers[32];
var byte PLsCount,VotetimeCountdown;

var() globalconfig byte MinVoters,ReqiuredMapsBeEnable,VotingTime;
var() globalconfig float MapChangePrc,MidGamePrc;
var() globalconfig bool bBeginEndGameVotingOnEndGame;
struct SaveInfoType
{
	var() globalconfig string MapName;
	var() globalconfig int GameIndex,PlayedTimes,LastPlayTime;
};
var() globalconfig SaveInfoType Maps[500];
struct SaveGIType
{
	var() globalconfig string GameName,ShortGameName,GamePrefix,GameTypeClass,Mutators,AddedCmdLine,Packages;
	var() globalconfig bool bKeepInventory;
};
var() globalconfig SaveGIType GameModes[10];
struct InitDataType
{
	var string MName;
	var int VotCount,GType,PTimes,LPTime,ActPos;
	var bool bInvalidMap;
};
var() globalconfig string AddedPackages[16],AddedPackageStr; // TEMP data
var InitDataType InitMaps[500];
var int NumVoteMaps;
var bool bVotingIsActive,bMessagedAndTried,bMapsInitilized,bMapListUpdatedNow;

function PostBeginPlay()
{
	local int i,j;
	local string CN;
	local bool bFoundErrors;
	
	Tag = 'BeginVoting'; // Allow some specific gametypes to trigger this to start voting.
	if( !bMapsInitilized )
		bMapsInitilized = True;
	else Return;

	// Check for configure error(s)
	CN = Caps(String(Class));
	For( i=0; i<10; i++ )
	{
		if( GameModes[i].GameName=="" )
			Continue;
		j = InStr(Caps(GameModes[i].Mutators),CN);
		if( j==-1 )
			Continue;
		Warn("Voting handler mutator was found on mutators list!");
		if( j==0 ) // First mutator on list
			GameModes[i].Mutators = Mid(GameModes[i].Mutators,Len(CN)+1);
		else GameModes[i].Mutators = Left(GameModes[i].Mutators,j-1)$Mid(GameModes[i].Mutators,j+Len(CN));
		bFoundErrors = True;
	}
	if( bFoundErrors )
		SaveConfig();
	// Initilize maplist
	For( i=0; i<500; i++ )
	{
		if( Maps[i].MapName=="" )
			Continue;
		if( Maps[i].GameIndex<0 )
		{
			For( j=0; j<10; j++ )
			{
				if( GameModes[j].GameName!="" && (GameModes[j].GamePrefix=="" ||
				 Left(Maps[i].MapName,Len(GameModes[j].GamePrefix))~=GameModes[j].GamePrefix) )
				{
					InitMaps[NumVoteMaps].MName = Maps[i].MapName;
					InitMaps[NumVoteMaps].PTimes = Maps[i].PlayedTimes;
					InitMaps[NumVoteMaps].VotCount = 0;
					InitMaps[NumVoteMaps].GType = j;
					InitMaps[NumVoteMaps].LPTime = Maps[i].LastPlayTime;
					InitMaps[NumVoteMaps].ActPos = i;
					InitMaps[NumVoteMaps].bInvalidMap = (Maps[i].PlayedTimes>0 && ReqiuredMapsBeEnable>Maps[i].LastPlayTime);
					NumVoteMaps++;
					if( NumVoteMaps==500 )
						Return;
				}
			}
		}
		else
		{
			InitMaps[NumVoteMaps].MName = Maps[i].MapName;
			InitMaps[NumVoteMaps].PTimes = Maps[i].PlayedTimes;
			InitMaps[NumVoteMaps].VotCount = 0;
			InitMaps[NumVoteMaps].GType = Maps[i].GameIndex;
			InitMaps[NumVoteMaps].LPTime = Maps[i].LastPlayTime;
			InitMaps[NumVoteMaps].ActPos = i;
			InitMaps[NumVoteMaps].bInvalidMap = (Maps[i].PlayedTimes>0 && ReqiuredMapsBeEnable>Maps[i].LastPlayTime);
			NumVoteMaps++;
			if( NumVoteMaps==500 )
				Return;
		}
	}
}
function bool CheckReplacement(Actor Other, out byte bSuperRelevant)
{
	if( Other.bIsPawn && PlayerPawn(Other)!=None )
	{
		UpdatingPlayers[PLsCount] = PlayerPawn(Other);
		PLsCount++;
		SetTimer(0.5,False);
	}
	return true;
}
function Timer()
{
	local int i;
	
	Tag = 'BeginVoting';
	For( i=0; i<PLsCount; i++ )
	{
		if( UpdatingPlayers[i]!=None && !UpdatingPlayers[i].bDeleteMe && UpdatingPlayers[i].Player!=None )
			AddVotingFor(UpdatingPlayers[i]);
		UpdatingPlayers[i] = None;
	}
	PLsCount = 0;
}
function AddVotingFor( PlayerPawn Other )
{
	local int i;
	
	For( i=0; i<MaxUnrealPlayers; i++ )
	{
		if( ActiveVReplications[i]==None || ActiveVReplications[i].bDeleteMe )
		{
			ActiveVReplications[i] = Spawn(Class'VotingReplicationInfo',Other);
			ActiveVReplications[i].TheVotingHandler = Self;
			ActiveVReplications[i].OpenReplication();
			Return;
		}
	}
}

function GetMapInfo( int ID, out string MName, out int GNum, out int NVot, out int NPInd, out int PTim, out byte bDis )
{
	MName = InitMaps[ID].MName;
	GNum = InitMaps[ID].GType;
	NVot = InitMaps[ID].VotCount;
	NPInd = InitMaps[ID].LPTime;
	PTim = InitMaps[ID].PTimes;
	if( InitMaps[ID].bInvalidMap )
		bDis = 1;
	else bDis = 0;
}
function GetGTInfo( byte ID, out string GName, out string SName, out string Prefix )
{
	GName = GameModes[ID].GameName;
	SName = GameModes[ID].ShortGameName;
	Prefix = GameModes[ID].GamePrefix;
}

event Trigger( Actor Other, Pawn EventInstigator )
{
	Level.Game.bGameEnded = True;
	// Disable gameinfo timing to stop it from auto-switching to next map
	Level.Game.SetTimer(0,False);
	Level.Game.Disable('Tick');
	Level.Game.SentText = -9999999;
	BeginGameVoting(False);
}

function BeginGameVoting( bool bMidGameVoting )
{
	local int i;
	
	if( bVotingIsActive )
		Return;
	Disable('Tick');
	bVotingIsActive = True;
	if( bMidGameVoting )
	{
		For( i=0; i<MaxUnrealPlayers; i++ )
		{
			if( ActiveVReplications[i]!=None && !ActiveVReplications[i].bDeleteMe )
				ActiveVReplications[i].MidGameVBegun();
		}
	}
	VotetimeCountdown = VotingTime;
	GoToState('VotingInProgress');
}

function ClientVoted( VotingReplicationInfo VRep, PlayerPawn VoteCaster, int MapIndex )
{
	local int i;
	
	if( MapIndex==VRep.CurrentlyVotedMap ) Return;
	if( MapIndex<0 || MapIndex>=NumVoteMaps )
	{
		VoteCaster.ClientMessage("Invalid map selection");
		Return;
	}
	if( InitMaps[MapIndex].bInvalidMap )
	{
		VoteCaster.ClientMessage("Your selected map is disabled");
		Return;
	}
	For( i=0; i<MaxUnrealPlayers; i++ )
	{
		if( ActiveVReplications[i]!=None && !ActiveVReplications[i].bDeleteMe )
			ActiveVReplications[i].PlayerVotedMap(VoteCaster.PlayerReplicationInfo,MapIndex);
	}
	if( VRep.CurrentlyVotedMap==-1 )
	{
		InitMaps[MapIndex].VotCount++;
		VRep.CurrentlyVotedMap = MapIndex;
		For( i=0; i<MaxUnrealPlayers; i++ )
			if( ActiveVReplications[i]!=None )
				ActiveVReplications[i].UpdateVotesCount(MapIndex,InitMaps[MapIndex].VotCount,-1);
	}
	else
	{
		InitMaps[MapIndex].VotCount++;
		InitMaps[VRep.CurrentlyVotedMap].VotCount--;
		For( i=0; i<MaxUnrealPlayers; i++ )
			if( ActiveVReplications[i]!=None )
				ActiveVReplications[i].UpdateVotesCount(MapIndex,InitMaps[MapIndex].VotCount,VRep.CurrentlyVotedMap,InitMaps[VRep.CurrentlyVotedMap].VotCount);
		VRep.CurrentlyVotedMap = MapIndex;
	}
	TallyVotes();
}

// Check for enough votes to force to that map or for midgame voting.
function TallyVotes()
{
	local int i,c;
	local float p,tp;
	
	For( i=0; i<MaxUnrealPlayers; i++ )
		if( ActiveVReplications[i]!=None )
			c++; // First count all active voters

	if( MinVoters>c )
		Return;
	
	For( i=0; i<NumVoteMaps; i++ )
	{
		if( InitMaps[i].VotCount==0 )
			Continue;
		p = (float(InitMaps[i].VotCount)/float(c)); // Count the percent of votes the map has
		tp+=p;
		if( p>MapChangePrc )
			VotingWon(i);
	}
	if( !bVotingIsActive && tp>MidGamePrc )
		BeginGameVoting(True);
}

// A mapvote has won!
function VotingWon( int MIndex )
{
	local int i;
	local string NewURL;
	
	Disable('Tick');
	bVotingIsActive = True;
	For( i=0; i<MaxUnrealPlayers; i++ )
		if( ActiveVReplications[i]!=None )
			ActiveVReplications[i].TurnOffVotingMenu();
	if( MIndex==-1 )
		MIndex = Rand(NumVoteMaps);
	if( !bMessagedAndTried )
	{
		For( i=0; i<MaxUnrealPlayers; i++ )
		{
			if( ActiveVReplications[i]!=None && !ActiveVReplications[i].bDeleteMe )
				ActiveVReplications[i].MapVotingWon(MIndex);
		}
		if( !bMapListUpdatedNow )
		{
			Maps[InitMaps[MIndex].ActPos].PlayedTimes++;
			Maps[InitMaps[MIndex].ActPos].LastPlayTime = -1;
			For( i=0; i<500; i++ )
			{
				if( Maps[i].MapName=="" )
					Continue;
				Maps[i].LastPlayTime++;
			}
			SaveConfig();
			bMessagedAndTried = True;
		}
	}
	NewURL = InitMaps[MIndex].MName$"?Game="$GameModes[InitMaps[MIndex].GType].GameTypeClass$"?Mutator="$string(Class);
	if( GameModes[InitMaps[MIndex].GType].Mutators!="" )
		NewURL = NewURL$","$GameModes[InitMaps[MIndex].GType].Mutators;
	if( GameModes[InitMaps[MIndex].GType].AddedCmdLine!="" )
		NewURL = NewURL$"?"$GameModes[InitMaps[MIndex].GType].AddedCmdLine;
	UpdateServerPacks(InitMaps[MIndex].GType);
	Level.ServerTravel(NewURL,GameModes[InitMaps[MIndex].GType].bKeepInventory);
	GoToState('SecurityChecking');
}

function UpdateServerPacks( byte GameMode ) // Add or remove any neccessary serverpackages here.
{
	local string S,Q,V;
	local int i,l;
	local byte j;

	if( int(Level.EngineVersion)<227 || AddedPackageStr==GameModes[GameMode].Packages )
		Return; // Not supported by older versions.
	Q = Chr(34); // Init " char.
	S = ConsoleCommand("Get GameEngine ServerPackages");

	// First remove all old packages last gametype used.
	for( j=0; j<ArrayCount(AddedPackages); j++ )
	{
		if( AddedPackages[j]=="" )
			Break;
		V = Q$AddedPackages[j]$Q;
		i = InStr(S,V);
		if( i==-1 )
			Continue;
		l = Len(V);
		if( Mid(S,i-1,1)=="," )
			S = Left(S,i-1)$Mid(S,i+Len(V));
		else if( Mid(S,i+l,1)=="," )
			S = Left(S,i)$Mid(S,i+l+1);
		else S = Left(S,i)$Mid(S,i+l); // Empty server packages.. shouldn't happen..
	}

	// Now add in new desired packages.
	AddedPackageStr = GameModes[GameMode].Packages;
	V = GameModes[GameMode].Packages;
	i = InStr(V,",");
	j = 0;
	While( True )
	{
		if( i==-1 )
		{
			AddedPackages[j++] = V;
			if( j<ArrayCount(AddedPackages) )
				AddedPackages[j] = "";
			Break;
		}
		AddedPackages[j++] = Left(V,i);
		if( j>=ArrayCount(AddedPackages) )
			Break;
		V = Mid(V,i+1);
		i = InStr(V,",");
	}
	for( j=0; j<ArrayCount(AddedPackages); j++ )
	{
		if( AddedPackages[j]=="" )
			Break;
		V = Q$AddedPackages[j]$Q;
		if( InStr(S,V)!=-1 )
			Continue;
		l = Len(S);
		if( l<=2 )
			S = "("$V$")";
		else S = Left(S,l-1)$","$V$")";
	}
	ConsoleCommand("Set GameEngine ServerPackages"@S);
	SaveConfig();
}

function ClientDisconnected( VotingReplicationInfo Other )
{
	local int i;
	
	For( i=0; i<MaxUnrealPlayers; i++ )
		if( ActiveVReplications[i]==Other )
		{
			ActiveVReplications[i] = None;
			Break;
		}
	if( Other.CurrentlyVotedMap!=-1 )
	{
		InitMaps[Other.CurrentlyVotedMap].VotCount--;
		For( i=0; i<MaxUnrealPlayers; i++ )
			if( ActiveVReplications[i]!=None )
				ActiveVReplications[i].UpdateVotesCount(Other.CurrentlyVotedMap,InitMaps[Other.CurrentlyVotedMap].VotCount,-1);
		Other.CurrentlyVotedMap = -1;
		TallyVotes();
	}
}

function ForceMapChange()
{
	local int i,c,bn;
	local float p,bp;
	
	For( i=0; i<MaxUnrealPlayers; i++ )
		if( ActiveVReplications[i]!=None )
			c++;
	
	bn = -1;
	For( i=0; i<NumVoteMaps; i++ )
	{
		if( InitMaps[i].VotCount==0 )
			Continue;
		p = (float(InitMaps[i].VotCount)/float(c));
		if( bn==-1 || p>bp )
		{
			bn = i;
			bp = p;
		}
	}
	VotingWon(bn);
}
function OpenVotingMenus()
{
	local int i;
	
	For( i=0; i<MaxUnrealPlayers; i++ )
		if( ActiveVReplications[i]!=None )
			ActiveVReplications[i].BringVotingMenu();
}

State VotingInProgress
{
Ignores Tick;

Begin:
	Sleep(Level.TimeDilation*2);
	OpenVotingMenus();
Timing:
	Sleep(Level.TimeDilation);
	if( VotetimeCountdown>0 )
		VotetimeCountdown--;
	if( VotetimeCountdown<=0 )
		ForceMapChange();
	else if( VotetimeCountdown>0 && VotetimeCountdown<=10 )
		BroadcastMessage(VotetimeCountdown$"..",True,'CriticalEvent');
	else if( VotetimeCountdown==15 )
		BroadcastMessage("15 seconds..",True,'CriticalEvent');
	else if( VotetimeCountdown==30 )
		BroadcastMessage("30 seconds left..",True,'CriticalEvent');
	else if( VotetimeCountdown==60 )
		BroadcastMessage("1 minute remains..",True,'CriticalEvent');
	GoTo'Timing';
}

State SecurityChecking
{
Ignores Tick;

Begin:
	Sleep(Level.TimeDilation*2);
	if( Level.NextSwitchCountdown<0 )
		VotingWon(-1);
	GoTo'Begin';
}

function Tick( float Delta )
{
	if( !bBeginEndGameVotingOnEndGame )
	{
		Disable('Tick');
		Return;
	}
	if( Level.Game.bGameEnded )
		Trigger(None,None);
}

function AddMapToList( VotingReplicationInfo Sender, string MName )
{
	local int i,free;
	
	free = -1;
	For( i=0; i<ArrayCount(Maps); i++ )
	{
		if( Maps[i].MapName=="" )
		{
			free = i;
			Break;
		}
	}
	if( free==-1 )
	{
		PlayerPawn(Sender.Owner).ClientMessage("Can't add to maplist; becouse maplist is full!");
		Return;
	}
	Maps[free].MapName = MName;
	Maps[free].GameIndex = -1;
	Maps[free].PlayedTimes = 0;
	Maps[free].LastPlayTime = 0;
	For( i=0; i<MaxUnrealPlayers; i++ )
	{
		if( ActiveVReplications[i]!=None && !ActiveVReplications[i].bDeleteMe && ActiveVReplications[i].FillingEntry>1 )
			ActiveVReplications[i].FillMapInfo(MName,0,free);
	}
	SaveConfig();
	bMapListUpdatedNow = True;
}
function SetMapGIndex( VotingReplicationInfo Sender, string MName, int GIndex )
{
	local int i,free,x;
	
	free = -1;
	For( i=0; i<ArrayCount(Maps); i++ )
	{
		if( Maps[i].MapName~=MName )
		{
			if( Maps[i].GameIndex==GIndex )
				Return;
			Maps[i].GameIndex = GIndex;
			Maps[i].LastPlayTime = 0;
			SaveConfig();
			bMapListUpdatedNow = True;

			For( x=0; x<MaxUnrealPlayers; x++ )//use x, not i... DOTS!!!!!!!!!!!!!!!!
			{
				if( ActiveVReplications[x]!=None && !ActiveVReplications[x].bDeleteMe && ActiveVReplications[x].FillingEntry>1 )
					ActiveVReplications[x].FillMapInfo(Maps[i].MapName,GIndex+1,i);
			}
	
			Return;
		}
		else if( free==-1 && Maps[i].MapName=="" )
			free = i;
	}
	PlayerPawn(Sender.Owner).ClientMessage("Can't change map gameindex; because map is not found on maplist!");
}
function RemoveMapFromList( VotingReplicationInfo Sender, string MName )
{
	local int i,j;
	
	For( i=0; i<ArrayCount(Maps); i++ )
	{
		if( Maps[i].MapName~=MName )
		{
			For( j=i+1; j<ArrayCount(Maps); j++ )
			{
				Maps[j-1].MapName = Maps[j].MapName;
				Maps[j-1].GameIndex = Maps[i].GameIndex;
				Maps[j-1].PlayedTimes = Maps[i].PlayedTimes;
				Maps[j-1].LastPlayTime = Maps[i].LastPlayTime;
			}
			Maps[ArrayCount(Maps)-1].MapName = "";
			Maps[ArrayCount(Maps)-1].GameIndex = -1;
			Maps[ArrayCount(Maps)-1].PlayedTimes = 0;
			Maps[ArrayCount(Maps)-1].LastPlayTime = 0;
			SaveConfig();
			bMapListUpdatedNow = True;
			For( i=0; i<MaxUnrealPlayers; i++ )
			{
				if( ActiveVReplications[i]!=None && !ActiveVReplications[i].bDeleteMe && ActiveVReplications[i].FillingEntry>1 )
				{
					ActiveVReplications[i].bFillingMapList = True;
					ActiveVReplications[i].FillingEntry = 2;
					if( ActiveVReplications[i].LastSentInfoX>=(i-1) )
					{
						ActiveVReplications[i].LastSentInfoX = i-1;
						if( ActiveVReplications[i].LastSentInfoX<=0 )
							ActiveVReplications[i].LastSentInfoX = 0;
					}
					ActiveVReplications[i].Enable('Tick');
				}
			}
			Return;
		}
	}
	PlayerPawn(Sender.Owner).ClientMessage("Can't remove from maplist; because map is not found in maplist!!!");
}
function GetGameTypeInfo( byte Index, out string GTypeName, out string MPrefix )
{
	GTypeName = GameModes[Index].GameName;
	MPrefix = GameModes[Index].GamePrefix;
}
function GetMapInfoAdmin( int Index, out string MName, out int GIndex )
{
	MName = Maps[Index].MapName;
	GIndex = Maps[Index].GameIndex;
}

// For use of webadmin access
static function GetFullActGTs( out string GNMS[10], out string ShrN[10], out string GPRF[10], out string GTCL[10], out string MTS[10], out string ACMD[10], out string Pcks[10] )
{
	local int i;

	For( i=0; i<10; i++ )
	{
		GNMS[i] = Default.GameModes[i].GameName;
		ShrN[i] = Default.GameModes[i].ShortGameName;
		GPRF[i] = Default.GameModes[i].GamePrefix;
		GTCL[i] = Default.GameModes[i].GameTypeClass;
		MTS[i] = Default.GameModes[i].Mutators;
		ACMD[i] = Default.GameModes[i].AddedCmdLine;
		Pcks[i] = Default.GameModes[i].Packages;
	}
}
static function Static_UpdFullGTs( int i, string GNMS, string ShrN, string GPRF, string GTCL, string MTS, string ACMD, string Pcks )
{
	Default.GameModes[i].GameName = GNMS;
	Default.GameModes[i].ShortGameName = ShrN;
	Default.GameModes[i].GamePrefix = GPRF;
	Default.GameModes[i].GameTypeClass = GTCL;
	Default.GameModes[i].Mutators = MTS;
	Default.GameModes[i].AddedCmdLine = ACMD;
	Default.GameModes[i].Packages = Pcks;
	StaticSaveConfig();
}
static function bool Static_IsAvailable( int Idx )
{
	Return (Default.GameModes[Idx].GameTypeClass!="");
}

static function GetFullActMaps( out string MName[500], out int GIdex[500], out int MLen )
{
	local int i;

	For( i=0; i<500; i++ )
	{
		if( Default.Maps[i].MapName=="" )
			Continue;
		MName[MLen] = Default.Maps[i].MapName;
		GIdex[MLen++] = Default.Maps[i].GameIndex;
	}
}
static function GetIdxActMaps( out string MName[500], int GIdex, out int MLen )
{
	local int i;

	For( i=0; i<500; i++ )
	{
		if( Default.Maps[i].GameIndex!=GIdex || Default.Maps[i].MapName=="" )
			Continue;
		MName[MLen++] = Default.Maps[i].MapName;
	}
}
static function Static_UpdFullActMaps( string MName[500], int GIdex[500] )
{
	local int i;

	For( i=0; i<500; i++ )
	{
		if( !(Default.Maps[i].MapName~=MName[i]) )
		{
			Default.Maps[i].MapName = MName[i];
			Default.Maps[i].PlayedTimes = 0;
			Default.Maps[i].LastPlayTime = 0;
		}
		Default.Maps[i].GameIndex = GIdex[i];
	}
	StaticSaveConfig();
}

function UpdateFromStatic()
{
	local int i;

	For( i=0; i<500; i++ )
	{
		Maps[i].MapName = Default.Maps[i].MapName;
		Maps[i].PlayedTimes = Default.Maps[i].PlayedTimes;
		Maps[i].LastPlayTime = Default.Maps[i].LastPlayTime;
		Maps[i].GameIndex = Default.Maps[i].GameIndex;
	}
	For( i=0; i<10; i++ )
	{
		GameModes[i].GameName = Default.GameModes[i].GameName;
		GameModes[i].ShortGameName = Default.GameModes[i].ShortGameName;
		GameModes[i].GamePrefix = Default.GameModes[i].GamePrefix;
		GameModes[i].GameTypeClass = Default.GameModes[i].GameTypeClass;
		GameModes[i].Mutators = Default.GameModes[i].Mutators;
		GameModes[i].AddedCmdLine = Default.GameModes[i].AddedCmdLine;
		GameModes[i].Packages = Default.GameModes[i].Packages;
	}
}

defaultproperties
{
				MinVoters=1
				ReqiuredMapsBeEnable=5
				VotingTime=40
				MapChangePrc=0.750000
				MidGamePrc=0.510000
}
